from typing import *

from lark import Lark

from .backend_base import BaseCombatBackend
from .combat_config_parser import CombatConfig, PriorityLine, get_sprinty_grammar, TreeToConfig
from ..sprinty_combat import SprintyCombat

class CombatConfigProvider(BaseCombatBackend):
    def __init__(self, path: str, cast_time: float = 0.2):
        super().__init__(cast_time=cast_time)
        self.filename = path
        with open(path) as file:
            self.config: CombatConfig = self.parse_config(file.read())

    def parse_config(self, file_contents) -> CombatConfig:
        grammar = get_sprinty_grammar()

        parser = Lark(grammar)
        tree = parser.parse(file_contents)
        return TreeToConfig().transform(tree)

    async def get_real_round(self, r: int) -> Optional[PriorityLine]:
        if r in self.config.specific_rounds:
            return self.config.specific_rounds[r]
        return None

    async def get_relative_round(self, r: int) -> Optional[PriorityLine]:
        if len(self.config.infinite_rounds) > 0:
            return self.config.infinite_rounds[r % len(self.config.infinite_rounds)]
        return None

    async def handle_no_cards_given(self):
        raise RuntimeError(f"Full config fail! \"{self.filename}\" might be empty or contains only explicit rounds. Consider adding a pass or something else")



from typing import *

from wizwalker import ClientHandler, Client

from .sprinty_client import SprintyClient


def upgrade_clients(clients: List[Client]) -> List[SprintyClient]:
    for client in clients:
        client.__class__ = SprintyClient
    return clients


class WizSprinter(ClientHandler):
    def __repr__(self):
        return f"<WizSprinter {self.clients=}>"

    def get_new_clients(self) -> List[SprintyClient]:
        """Upgraded get_new_clients method that returns SprintyClients"""
        return upgrade_clients(super().get_new_clients())

    def remove_dead_clients(self) -> List[SprintyClient]:
        """
        Remove and return clients that are no longer running

        Returns:
            List of the dead clients removed
        """
        return upgrade_clients(super().remove_dead_clients())

    def get_ordered_clients(self) -> List[SprintyClient]:
        """
        Get clients ordered by their position on the screen

        Returns:
            List of the ordered clients
        """
        return upgrade_clients(super().get_ordered_clients())



from typing import *

from wizwalker.combat.member import CombatMember

from .backend_base import BaseCombatBackend
from .combat_api import *
from ..sprinty_combat import SprintyCombat

class CombatConfigGenerator(BaseCombatBackend):
    def __init__(self, cast_time: float = 0.2):
        super().__init__(cast_time=cast_time)
        self.combat: SprintyCombat # restore autocomplete

    async def get_heal_priorities(self) -> list[MoveConfig]:
        me = await self.combat.get_client_member()
        client_health = await me.health()
        client_max_health = await me.max_health() # TODO: Can this be zero?
        if client_health / client_max_health > 0.4:
            return []

        return [
            MoveConfig(
                move=Move(
                    card=TemplateSpell(requirements=[SpellType.type_heal_self]),
                )
            ),
            MoveConfig(
                move=Move(
                    card=TemplateSpell(requirements=[SpellType.type_heal, SpellType.type_aoe]),
                )
            ),
        ]

    async def get_real_round(self, r: int) -> Optional[PriorityLine]:
        priorities = []

        # heal
        priorities.extend(await self.get_heal_priorities())
        
        # attack, aoe
        priorities.append(
            MoveConfig(
                move=Move(
                    card=TemplateSpell(requirements=[SpellType.type_damage, SpellType.type_aoe]),
                    enchant=TemplateSpell(requirements=[SpellType.type_enchant], optional=True)
                )
            )
        )

        # attack
        priorities.append(
            MoveConfig(
                move=Move(
                    card=TemplateSpell(requirements=[SpellType.type_damage]),
                    enchant=TemplateSpell(requirements=[SpellType.type_enchant], optional=True)
                ),
                target=TargetData(TargetType.type_enemy)
            )
        )

        # pass
        priorities.append(            
            MoveConfig(move=Move(
                card=NamedSpell("pass")
            ))
        )

        # TODO: Maybe dedupe?
        return PriorityLine(priorities=priorities)

    async def get_relative_round(self, r: int) -> Optional[PriorityLine]:
        return self.get_real_round(r) # it's all automatic, so they mean the same thing


from enum import Enum, auto
from typing import *

class TargetType(Enum):
    type_self = auto()
    type_boss = auto()
    type_enemy = auto()
    type_ally = auto()
    type_aoe = auto()
    type_named = auto()


class SpellType(Enum):
    type_damage = auto()
    type_aoe = auto()
    type_heal = auto()
    type_heal_self = auto()
    type_heal_other = auto()
    type_blade = auto()
    type_shield = auto()
    type_trap = auto()
    type_enchant = auto()


class Spell:
    pass


class NamedSpell(Spell):
    def __init__(self, name: str, is_literal: bool = False):
        self.name = name
        self.is_literal = is_literal

    def __repr__(self) -> str:
        return f"NamedSpell(name=\"{self.name}\", is_literal={self.is_literal})"


class TemplateSpell(Spell):
    def __init__(self, requirements: List[SpellType], optional=False) -> None:
        self.requirements = requirements
        self.optional = optional

    def __repr__(self) -> str:
        return f"TemplateSpell(requirements={self.requirements}, optional={self.optional})"


class Move:
    def __init__(self, card: Spell, enchant: Spell = None):
        self.card = card
        self.enchant = enchant

    def __repr__(self) -> str:
        return f"Move(card={self.card}, enchant={self.enchant})"


class TargetData:
    def __init__(self, target_type: TargetType, extra_data: Any = None, is_literal: bool = False):
        self.target_type = target_type
        self.extra_data = extra_data
        self.is_literal = is_literal

    def __repr__(self) -> str:
        return f"TargetData(target_type={self.target_type}, extra_data={self.extra_data}, is_literal={self.is_literal})"


class MoveConfig:
    def __init__(self, move: Move, target: TargetData = None):
        self.move = move
        self.target = target

    def __repr__(self) -> str:
        return f"MoveConfig(move={self.move}, target={self.target})"


class PriorityLine:
    def __init__(self, priorities: List[MoveConfig], _round: int = None):
        self.priorities = priorities
        self.round = _round

    def __repr__(self) -> str:
        return f"PriorityLine(priorities={self.priorities}, round={self.round})"


class CombatConfig:
    def __init__(self, rounds: List[PriorityLine]):
        self.specific_rounds: dict[int, PriorityLine] = {}
        self.infinite_rounds: list[PriorityLine] = []
        for _round in rounds:
            if _round.round is None:
                self.infinite_rounds.append(_round)
            else:
                self.specific_rounds[_round.round] = _round

    def __repr__(self) -> str:
        return f"CombatConfig(specific_rounds={self.specific_rounds}, infinite_rounds={self.infinite_rounds})"


import asyncio
from typing import *

import wizwalker
from wizwalker.combat import CombatHandler
from wizwalker.combat import CombatMember
from wizwalker.combat.card import CombatCard
from wizwalker.memory import EffectTarget, SpellEffects

from .combat_backends.combat_config_parser import TargetType, TargetData, MoveConfig, TemplateSpell \
    , NamedSpell, SpellType, Spell
from .combat_backends.backend_base import BaseCombatBackend


class SprintyCombat(CombatHandler):
    def __init__(self, client: wizwalker.client.Client, config_provider: BaseCombatBackend):
        super().__init__(client)
        self.client: wizwalker.client.Client = client # to restore autocomplete
        self.config = config_provider
        self.turn_adjust = 0
        self.cur_card_count = 0
        self.prev_card_count = 0
        self.was_pass = False
        self.had_first_round = False
        self.rel_round_offset = 0

    async def handle_combat(self):
        self.turn_adjust = 0
        self.cur_card_count = 0
        self.prev_card_count = 0
        self.rel_round_offset = 0
        self.was_pass = False
        self.had_first_round = False
        await super().handle_combat()

    async def get_member_named(self, name: str) -> Optional[CombatMember]:
        # Issue: #4
        async def _inner():
            members: List[CombatMember] = await self.get_members()

            for member in members:
                if name == await member.name():
                    return member
            return None
        try:
            return await wizwalker.utils.maybe_wait_for_value_with_timeout(
                _inner,
                timeout=2.0
            )
        except wizwalker.errors.ExceptionalTimeout:
            return None

    async def get_member_vaguely_named(self, name: str) -> Optional[CombatMember]:
        # Issue #4
        async def _inner():
            members = await self.get_members()

            for member in members:
                if name.lower() in (await member.name()).lower():
                    return member
            return None
        try:
            return await wizwalker.utils.maybe_wait_for_value_with_timeout(
                _inner,
                timeout=2.0
            )
        except wizwalker.errors.ExceptionalTimeout:
            return None

    async def pass_button(self):
        self.was_pass = True
        await super().pass_button()

    async def get_cards(self) -> List[CombatCard]:  # extended to sort by enchanted
        async def _inner() -> List[CombatCard]:
            cards = await super(SprintyCombat, self).get_cards()
            rese, res = [], []
            for card in cards:
                if await card.is_enchanted():
                    rese.append(card)
                else:
                    res.append(card)
            return rese + res
        try:
            return await wizwalker.utils.maybe_wait_for_any_value_with_timeout(_inner, sleep_time=0.2, timeout=2.0)
        except wizwalker.errors.ExceptionalTimeout:
            return []

    async def get_card_named(self, name: str) -> Optional[CombatCard]:
        try:
            return await super().get_card_named(name)
        except ValueError:
            return None

    async def get_card_with_predicate(self, pred: Callable) -> Optional[CombatCard]:
        cards = await self.get_cards_with_predicate(pred)
        if len(cards) > 0:
            return cards[0]
        return None

    async def get_card_vaguely_named(self, name: str) -> Optional[CombatCard]:
        async def _pred(card: CombatCard):
            return name.lower() in (await card.name()).lower()

        return await self.get_card_with_predicate(_pred)

    async def get_card_counts(self) -> Tuple[int, int]:
        # Issue: #6. Very rare error
        async def _inner():
            window = None
            while window is None:
                window, *_ = await self.client.root_window.get_windows_with_name("CountText")
            text: str = await window.maybe_text()
            _, count_text = text.splitlines()
            count_text = count_text[8:-9]
            count_text = count_text.replace("of", "").strip()  # I know this sucks
            res1, res2 = count_text.split()
            return int(res1), int(res2)
        try:
            return await wizwalker.utils.maybe_wait_for_any_value_with_timeout(_inner, sleep_time=0.2, timeout=2.0)
        except wizwalker.errors.ExceptionalTimeout:
            return (0, 0) # TODO: Maybe propagate, but good enough for now

    async def get_castable_cards(self) -> List[CombatCard]:  # extension for castable cards only
        async def _pred(card: CombatCard):
            return await card.is_castable()

        return await self.get_cards_with_predicate(_pred)

    async def get_castable_cards_named(self, name: str) -> List[CombatCard]:
        cards = await self.get_castable_cards()
        res = []

        for card in cards:
            if name == await card.name():
                res.append(card)

        return res

    async def get_castable_cards_vaguely_named(self, name: str) -> List[CombatCard]:
        cards = await self.get_castable_cards()
        res = []
        for card in cards:
            if name.lower() in (await card.name()).lower():
                res.append(card)

        return res

    async def get_castable_card_named(self, name: str, only_enchants=False) -> Optional[CombatCard]:  # extension to get only castable card
        cards = await self.get_castable_cards()

        for card in cards:
            if name == await card.name():
                if only_enchants:
                    for e in await card.get_spell_effects():
                        if await e.effect_target() is EffectTarget.spell:
                            return card
                    else:
                        continue
                return card

        return None

    async def get_castable_card_vaguely_named(self, name: str, only_enchants=False) -> Optional[CombatCard]:
        cards = await self.get_castable_cards()

        for card in cards:
            if name.lower() in (await card.name()).lower():
                if only_enchants:
                    for e in await card.get_spell_effects():
                        if await e.effect_target() is EffectTarget.spell:
                            return card
                    else:
                        continue
                return card

        return None

    async def get_castable_enchanted_card_named(self, name: str) -> Optional[CombatCard]:
        for s in await self.get_castable_cards_named(name):
            if await s.is_enchanted():
                return s
        return None

    async def get_castable_enchanted_card_vaguely_named(self, name: str) -> Optional[CombatCard]:
        for s in await self.get_castable_cards_vaguely_named(name):
            if await s.is_enchanted():
                return s
        return None

    async def get_cards_by_template(self, template: TemplateSpell) -> list[CombatCard]:
        cards = await self.get_castable_cards()
        res = []
        allow_aoe = SpellType.type_aoe in template.requirements
        for c in cards:
            fits = True
            effects = await c.get_spell_effects()
            c_type = (await c.type_name()).lower()
            for req in template.requirements:
                if req is SpellType.type_damage:
                    for e in effects:
                        target = await e.effect_target()
                        eff = await e.effect_type()
                        if target in (EffectTarget.enemy_team,
                                      EffectTarget.enemy_team_all_at_once,
                                      EffectTarget.enemy_single) \
                                and eff in (SpellEffects.damage,
                                            SpellEffects.damage_over_time,
                                            SpellEffects.damage_per_total_pip_power) \
                                or c_type == "damage":
                            if target in (EffectTarget.enemy_team, EffectTarget.enemy_team_all_at_once) and allow_aoe:
                                break
                            elif target in (EffectTarget.enemy_team, EffectTarget.enemy_team_all_at_once):
                                continue
                            else:
                                break
                    else:
                        fits = False
                elif req is SpellType.type_aoe:
                    for e in effects:
                        target = await e.effect_target()
                        if target in (EffectTarget.enemy_team, EffectTarget.enemy_team_all_at_once,
                                      EffectTarget.friendly_team, EffectTarget.friendly_team_all_at_once):
                            break
                    else:
                        fits = False
                elif req is SpellType.type_trap:
                    for e in effects:
                        target = await e.effect_target()
                        et = await e.effect_type()
                        if et is SpellEffects.modify_incoming_damage and target is EffectTarget.enemy_single:
                            break
                    else:
                        fits = False
                elif req is SpellType.type_shield:
                    for e in effects:
                        target = await e.effect_target()
                        et = await e.effect_type()
                        if et is SpellEffects.modify_incoming_damage and target is EffectTarget.friendly_single:
                            break
                    else:
                        fits = False
                elif req is SpellType.type_blade:
                    for e in effects:
                        target = await e.effect_target()
                        et = await e.effect_type()
                        if et is SpellEffects.modify_incoming_damage and target is EffectTarget.friendly_single:
                            break
                    else:
                        fits = False
                elif req is SpellType.type_heal:
                    for e in await c.get_spell_effects():
                        if (await e.effect_type()) is SpellEffects.heal:
                            break
                    else:
                        fits = False
                elif req is SpellType.type_heal_other:
                    for e in effects:
                        target = await e.effect_target()
                        if (target is EffectTarget.friendly_team or target is EffectTarget.friendly_single) and (
                                await e.effect_type()) is SpellEffects.heal:
                            break
                    else:
                        fits = False
                elif req is SpellType.type_heal_self:
                    for e in effects:
                        target = await e.effect_target()
                        if (target is EffectTarget.self or target is EffectTarget.friendly_team) and (
                                await e.effect_type()) is SpellEffects.heal:
                            break
                    else:
                        fits = False
                elif req is SpellType.type_enchant:
                    for e in effects:
                        if await e.effect_target() is EffectTarget.spell:
                            break
                    else:
                        fits = False

                if not fits:
                    break
            if fits:
                res.append(c)
        return res

    async def get_boss_or_none(self) -> Optional[CombatMember]:
        for m in await self.get_members():
            if await m.is_boss():
                return m
        return None

    async def get_allies(self) -> List[CombatMember]:
        members = []
        my_client = await self.get_client_member()
        my_participant = await my_client.get_participant()
        my_team_id = await my_participant.team_id()
        my_id = await my_participant.owner_id_full()
        for mem in await self.get_members():
            participant = await mem.get_participant()
            if await participant.team_id() == my_team_id \
                    and await participant.owner_id_full() != my_id:
                members.append(mem)
        return members

    async def get_enemies(self) -> List[CombatMember]:
        members = []
        my_client = await self.get_client_member()
        my_participant = await my_client.get_participant()
        my_team_id = await my_participant.team_id()
        for mem in await self.get_members():
            participant = await mem.get_participant()
            if await participant.team_id() != my_team_id:
                members.append(mem)
        return members

    async def get_nth_ally_or_none(self, n: int) -> Optional[CombatMember]:
        allies = await self.get_allies()
        if len(allies) <= n:
            return None
        return allies[n]

    async def get_nth_enemy_or_none(self, n: int) -> Optional[CombatMember]:
        enemies = await self.get_enemies()
        if len(enemies) <= n:
            return None
        return enemies[n]

    async def try_get_config_target(self, target: TargetData) -> Union[bool, Optional[CombatMember]]:
        ttype = None
        data = None
        if target is not None:
            ttype = target.target_type
            data = target.extra_data
        else:
            return None

        if ttype is TargetType.type_boss:
            if boss := await self.get_boss_or_none():
                return boss
        elif ttype is TargetType.type_self:
            return await self.get_client_member()
        elif ttype is TargetType.type_aoe:
            return None
        elif ttype is TargetType.type_enemy:
            if data is None:
                if enemy := await self.get_nth_enemy_or_none(0):
                    return enemy
            else:
                if enemy := await self.get_nth_enemy_or_none(data):
                    return enemy
        elif ttype is TargetType.type_ally:
            if data is None:
                if ally := await self.get_nth_ally_or_none(0):
                    return ally
            else:
                if ally := await self.get_nth_ally_or_none(data):
                    return ally
        elif ttype is TargetType.type_named:
            if target.is_literal:
                if res := await self.get_member_named(data):
                    return res
            if res := await self.get_member_vaguely_named(data):
                return res

        return False

    async def try_get_spell(self, spell: Spell, only_enchants=False) -> Union[CombatCard, str, None]:
        if isinstance(spell, NamedSpell):
            spell: NamedSpell
            if spell.name in ("pass", "none"):
                return spell.name
            if spell.is_literal:
                return await self.get_castable_card_named(spell.name, only_enchants)
            else:
                return await self.get_castable_card_vaguely_named(spell.name, only_enchants)
        elif isinstance(spell, TemplateSpell):
            spell: TemplateSpell
            res = await self.get_cards_by_template(spell)
            if len(res) > 0:
                return res[0]
            return None
        else:
            raise NotImplementedError("Unknown spell config type")

    async def try_execute_config(self, move_config: MoveConfig) -> bool:
        cur_card = await self.try_get_spell(move_config.move.card)
        if cur_card is None:
            return False

        if cur_card == "pass":
            await self.pass_button()
            return True

        target = await self.try_get_config_target(move_config.target)

        if target == False:  # Wouldn't want a None to mess it up
            return False

        if move_config.move.enchant is not None and not await cur_card.is_enchanted():
            enchant_card = await self.try_get_spell(move_config.move.enchant, only_enchants=True)
            if enchant_card != "none":
                if enchant_card is not None:
                    # Issue: 5. Casting wasn't that reliable
                    pre_enchant_count = len(await self.get_cards())
                    while len(await self.get_cards()) == pre_enchant_count:
                        await enchant_card.cast(cur_card, sleep_time=self.config.cast_time*2)
                        await asyncio.sleep(self.config.cast_time*2) # give it some time for card list to update

                    self.cur_card_count -= 1

                elif enchant_card is None and (isinstance(move_config.move.enchant, TemplateSpell) and not move_config.move.enchant.optional):
                    return False

        to_cast = await self.try_get_spell(move_config.move.card)
        if to_cast is None:
            return False  # this should not happen
        
        # Issue: 5. Casting wasn't that reliable
        try:
            while to_cast != None:
                try: 
                    await to_cast.cast(target, sleep_time=self.config.cast_time)
                    await asyncio.sleep(self.config.cast_time) # give it some time for card list to update
                    to_cast = await self.try_get_spell(move_config.move.card)
                except ValueError:
                    break # Issue: 8
        except wizwalker.errors.WizWalkerMemoryError or ValueError:
            pass # Let it happen if it happens
        return True

    async def fail_turn(self):
        self.turn_adjust -= 1
        await self.pass_button()

    async def on_fizzle(self):
        self.turn_adjust -= 1

    async def handle_round(self):
        try:
            await self.client.mouse_handler.activate_mouseless()
        except wizwalker.errors.HookAlreadyActivated:
            pass
        
        try:
            self.config.attach_combat(self) # For safety. Could probably also do this in handle_combat

            real_round = await self.round_number()
            self.cur_card_count = len(await self.get_cards()) + (await self.get_card_counts())[0]

            if not self.had_first_round:
                current_round = real_round - 1
                if current_round > 0:
                    self.turn_adjust -= current_round
            else:
                if self.cur_card_count >= self.prev_card_count and not self.was_pass:
                    await self.on_fizzle()

            self.was_pass = False
            current_round = (real_round - 1 + self.turn_adjust + self.rel_round_offset)

            # Issue: #3. Need to make sure it's valid
            member: CombatMember = None
            try:
                member = await wizwalker.utils.maybe_wait_for_any_value_with_timeout(
                    self.get_client_member,
                    timeout=2.0
                )
            except wizwalker.errors.ExceptionalTimeout:
                # TODO: Maybe make this more dramatic
                await self.fail_turn() # This is quite catastrophic. Use default fail for now
            
            if member is not None:
                if await member.is_stunned():
                    await self.fail_turn()
                else:
                    round_config = await self.config.get_real_round(real_round)
                    if round_config is None:
                        round_config = await self.config.get_relative_round(current_round)
                    else:
                        self.rel_round_offset -= 1

                    if round_config is not None:
                        for p in round_config.priorities:  # go through rounds priorities
                            if await self.try_execute_config(p):
                                break  # we found a working priority and managed to cast it
                        else:
                            await self.pass_button()
                    else:  # Very bad. Probably using empty config
                        await self.config.handle_no_cards_given()

            self.had_first_round = True  # might go bad on throw
            self.prev_card_count = self.cur_card_count
        finally:
            try:
                await self.client.mouse_handler.deactivate_mouseless()
            except wizwalker.errors.HookNotActive:
                pass



from wizwalker.combat import CombatCard
from wizwalker.memory import DynamicSpellEffect, SpellEffects

async def dump_card_effect(effect: DynamicSpellEffect, indent: int = 0):
    def printi(s):
        print(f"{' '*indent*4}{s}")
    
    printi(f"type_name: {await effect.read_type_name()}")
    indent += 1

    printi(f"effect_type: {await effect.effect_type()}")
    printi(f"effect_target: {await effect.effect_target()}")
    printi(f"effect_param: {await effect.effect_param()}")
    printi(f"act: {await effect.act()}")
    printi(f"act_num: {await effect.act_num()}")
    printi(f"converted: {await effect.converted()}")
    printi(f"damage_type: {await effect.string_damage_type()}")
    printi(f"disposition: {await effect.disposition()}")
    printi(f"num_rounds: {await effect.num_rounds()}")
    printi(f"param_per_round: {await effect.param_per_round()}")
    printi(f"pip_num: {await effect.pip_num()}")
    printi(f"protected: {await effect.protected()}")
    printi(f"enchantment_spell_template_id: {await effect.enchantment_spell_template_id()}")
    printi(f"rank: {await effect.rank()}")
    printi(f"chance_per_target: {await effect.chance_per_target()}")
    printi(f"armor_piercing_param: {await effect.armor_piercing_param()}")
    printi(f"cloaked: {await effect.cloaked()}")
    
    if await effect.effect_type() == SpellEffects.invalid_spell_effect:
        for sub in await effect.maybe_effect_list():
            await dump_card_effect(sub, indent)

async def dump_card_data(card: CombatCard):
    indent = 0
    def printi(s):
        print(f"{' '*indent*4}{s}")
    
    printi(f"{await card.name()} - {await card.display_name()}")
    indent += 1
    printi(f"type_name: {await card.type_name()}")
    printi(f"accuracy: {await card.accuracy()}")
    printi(f"is_item_card: {await card.is_item_card()}")
    printi(f"is_treasure_card: {await card.is_treasure_card()}")
    printi(f"is_side_board: {await card.is_side_board()}")
    printi(f"is_pve_only: {await card.is_pve_only()}")

    printi("effects:")
    for eff in await card.get_spell_effects():
        await dump_card_effect(eff, indent=indent+1)

class CombatLiftingEnvironment:

    async def lift_card_effects(self, card: CombatCard):
        pass

    async def calculate_card_score(self, card: CombatCard) -> int:
        # TODO: Make sure card is actually safe to use (memory valid)
        pass



from .wiz_sprinter import WizSprinter
from .sprinty_client import SprintyClient
from .sprinty_combat import SprintyCombat
from .combat_backends.config_backend import CombatConfigProvider
from .combat_backends.generator_backend import CombatConfigGenerator # workaround for circular imports



import asyncio
from wizwalker import XYZ, ClientHandler
from wizwalker import Keycode
from loguru import logger
from itertools import takewhile
import math
from wizwalker.memory import Window

handleQuestCollection = True
is_tab = '\t'.__eq__


@logger.catch()
async def go_through_dialog(p):
    while await p.is_in_dialog():
        await p.send_key(Keycode.SPACEBAR, 0.1)

    await asyncio.sleep(.5)


@logger.catch()
async def gateTypeDifferentiation(x, y, z, p, zoneAccessType):
    # standard zone gate, just teleport
    if zoneAccessType == 'standard':
        await p.teleport(XYZ(float(x), float(y), float(z)))

        await p.wait_for_zone_change()

    # dungeon ENTRANCE, teleport, click x, wait for change
    elif zoneAccessType == 'dungeon':

        await p.teleport(XYZ(float(x), float(y), float(z)))

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.8)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    # this is a dungeon exit, user may need to confirm to leave the dungeon.  Teleport, click confirm
    elif zoneAccessType == 'dungeonExitConfirm':

        if await p.is_in_dialog():
            await go_through_dialog(p)

        await p.teleport(XYZ(float(x), float(y), float(z)))

        await p.mouse_handler.activate_mouseless()
        await asyncio.sleep(1)

        try:
            await p.mouse_handler.click_window_with_name('centerButton')
            await p.wait_for_zone_change()
        except ValueError:
            await asyncio.sleep(8)

        await p.mouse_handler.deactivate_mouseless()

    elif zoneAccessType == 'xNoWait':

        if await p.is_in_dialog():
            await go_through_dialog(p)

        await p.teleport(XYZ(float(x), float(y), float(z)))

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xNoWaitMirageCaterwaulToCaravan':
        if await p.is_in_dialog():
            await go_through_dialog(p)

        await p.teleport(XYZ(float(x), float(y), float(z)))

        await asyncio.sleep(7)

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xNoWaitPolaris':
        await p.teleport(XYZ(float(x), float(y), float(z)))

        await asyncio.sleep(4)

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xSkipRideKrok1':
        await p.teleport(XYZ(4521.9609375, 3189.564208984375, 25.792266845703125))

        # there is a vendor near the boat, sleep to make sure you click the boat instead of the vendor
        await asyncio.sleep(1.5)

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xSkipRideKrok2':
        await p.teleport(XYZ(11749.1123046875, -189.94265747070312, 1219.797119140625))

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xNoWaitSkipRideMarleyboneChelsea':
        await p.teleport(XYZ(float(x), float(y), float(z)))

        await p.wait_for_zone_change()

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xNoWaitSkipRideMarleyboneChelseaReturn':
        await p.teleport(XYZ(float(x), float(y), float(z)))

        await p.wait_for_zone_change()

        await asyncio.sleep(.3)
        await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xNoWaitSkipRideMarleyboneHyde':
        await p.teleport(XYZ(float(x), float(y), float(z)))

        await p.wait_for_zone_change()

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await asyncio.sleep(4)

    elif zoneAccessType == 'xSkipRideMarleyboneIronworksReturn':
        await p.teleport(XYZ(float(x), float(y), float(z)))
        await asyncio.sleep(1)

        await p.mouse_handler.activate_mouseless()

        try:
            await p.mouse_handler.click_window_with_name('centerButton')
        except ValueError:
            await asyncio.sleep(0.01)

        await p.wait_for_zone_change()

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

        await p.mouse_handler.deactivate_mouseless()

    elif zoneAccessType == 'xNoWaitSkipRideMarleyboneHydeReturn':
        await p.teleport(XYZ(float(x), float(y), float(z)))

        await p.wait_for_zone_change()

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'dungeonSkipRideMarleyboneIronworks':
        await asyncio.sleep(1)

        await p.teleport(XYZ(float(x), float(y), float(z)))
        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'xSkipRideMarleyboneIronworksReturn':
        await p.send_key(Keycode.PAGE_UP, 0.1)
        await p.wait_for_zone_change()


    # special case, krokotopia obelisks
    elif zoneAccessType == 'xNoWaitKrokObelisk':
        # Obelisk 1 (to Tomb of Storms)
        await p.teleport(XYZ(-3287.2294921875, -2826.498779296875, -35.353118896484375))
        while not await p.is_in_npc_range():
            pass

        # print('sent x press for obelisk 1')
        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)

        await asyncio.sleep(.5)

        # Obelisk 2 (to Tomb of Storms)
        await p.teleport(XYZ(-4532.013671875, -2590.87451171875, -35.353668212890625))
        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
        await asyncio.sleep(.5)

        # Obelisk 3 (to Tomb of Storms)
        await p.teleport(XYZ(-4274.5419921875, -1374.5045166015625, -35.353607177734375))
        while not await p.is_in_npc_range():
            pass

        await asyncio.sleep(.5)
        while await p.is_in_npc_range():
            await asyncio.sleep(.4)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
            await p.send_key(Keycode.X, 0.1)
        logger.info("Waiting a long time for the mouth thing to open")
        await asyncio.sleep(20)
        logger.info("If you haven't changed zones by now, manually turn on the obelisks and walk to the zone gate.  The script should continue after you change zones")

        # Zone Change - Tomb of Storms
        # Teleport to front of mouth thing
        await p.teleport(XYZ(-3364.827392578125, -1802.46630859375, -35.354522705078125))

        await p.wait_for_zone_change()

    elif zoneAccessType == 'dungeonDragonSpireGrandChasm':
        await asyncio.sleep(1)
        await p.send_key(Keycode.PAGE_DOWN, 0.1)

        await p.teleport(XYZ(float(x), float(y), float(z)))

        await asyncio.sleep(.5)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.wait_for_zone_change()

        await asyncio.sleep(.3)

        await p.teleport(XYZ(1.292851448059082, 227.6787567138672, 24.999969482421875))
        await asyncio.sleep(1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)


        #await asyncio.sleep(10)
        #await p.teleport(XYZ(-4138.177734375, 4193.40185546875, 30.1983642578125))
        await asyncio.sleep(1)

    elif zoneAccessType == 'dungeonExitConfirmMana':
        await p.send_key(Keycode.PAGE_UP, 0.1)
        await p.send_key(Keycode.PAGE_UP, 0.1)
        await p.send_key(Keycode.PAGE_UP, 0.1)
        await p.wait_for_zone_change()


    elif zoneAccessType == 'xSkipRideDragonspireRoost':
        await p.send_key(Keycode.PAGE_DOWN)

        await p.teleport(XYZ(float(x), float(y), float(z)))

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await p.send_key(Keycode.X, 0.1)
            await asyncio.sleep(.4)

        await p.wait_for_zone_change()

        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)

        await asyncio.sleep(.5)

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await p.send_key(Keycode.X, 0.1)
            await asyncio.sleep(.4)

        await p.wait_for_zone_change()

        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)

        await asyncio.sleep(.5)

        await asyncio.sleep(1)

    elif zoneAccessType == 'xNoWaitDragonSpireReturnToAcademy':
        await p.send_key(Keycode.PAGE_UP, 0.1)

        await p.wait_for_zone_change()

    elif zoneAccessType == 'dungeonExitConfirmCelestiaTemple':

        if await p.is_in_dialog():
            await go_through_dialog(p)

        await asyncio.sleep(.3)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)
        await p.send_key(Keycode.SPACEBAR, 0.1)

        await asyncio.sleep(1)

        await p.teleport(XYZ(float(x), float(y), float(z)))

        await p.mouse_handler.activate_mouseless()
        await asyncio.sleep(1)

        try:
            await p.mouse_handler.click_window_with_name('centerButton')
        except ValueError:
            await asyncio.sleep(0.01)

        await p.mouse_handler.deactivate_mouseless()

        await p.wait_for_zone_change()

    elif zoneAccessType == 'khrysDungeon1':
        await p.teleport(XYZ(float(x), float(y), float(z)))

        await asyncio.sleep(2)

        while not await p.is_in_npc_range():
            pass

        while await p.is_in_npc_range():
            await p.send_key(Keycode.X, 0.1)
            await asyncio.sleep(.4)

        await p.wait_for_zone_change()
        # teleport to second zone door
        await p.teleport(XYZ(1647.79248046875, 29.44374656677246, 6.103515625e-05))
        await p.wait_for_zone_change()
        await asyncio.sleep(2)


    elif zoneAccessType == 'khrysSerpentIsland':
        await p.teleport(XYZ(float(x), float(y), float(z)))
        await p.wait_for_zone_change()
        await asyncio.sleep(2)

        await p.teleport(XYZ(1647.79248046875, 29.44374656677246, 6.103515625e-05))
        await asyncio.sleep(1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await p.send_key(Keycode.X, 0.1)
        await asyncio.sleep(4)

        await p.teleport(XYZ(7662.78564453125, 7625.587890625, 1265.4591064453125))
        await asyncio.sleep(.6)


# uses an interactive teleporter (such as those in empyrea and later worlds) to teleport between zones
@logger.catch()
async def interactiveTeleportToZone(p, menuButtonNumber):

    while not await p.is_in_npc_range():
        pass

    while await p.is_in_npc_range():
        await p.send_key(Keycode.X, 0.1)
        await asyncio.sleep(.4)

    await asyncio.sleep(.4)
    await p.mouse_handler.activate_mouseless()

    # 4 buttons per page - menuButtonNum / 4 rounded up to nearest whole number equals the page that the button is on
    actualButtonToClick = menuButtonNumber

    if menuButtonNumber > 4:
        pageNum = int(math.ceil(menuButtonNumber / 4)) - 1
        # click to correct page
        if pageNum > 0:
            for i in range(pageNum):
                await p.mouse_handler.click_window_with_name('rightButton')
                await asyncio.sleep(0.4)

        # the actual button number on the page, considering that there are 4 buttons per page (0 -> 3)
        actualButtonToClick = ((menuButtonNumber) - ((pageNum) * 4))

    await p.mouse_handler.click_window_with_name('opt' + str(actualButtonToClick - 1))
    await asyncio.sleep(.4)
    await p.mouse_handler.click_window_with_name('teleportButton')
    await p.wait_for_zone_change()

    await p.mouse_handler.deactivate_mouseless()


@logger.catch()
async def createStack(p1WorldName):

    lines = await parseFile("traversalData/zoneMap.txt", p1WorldName)

    lines = iter(lines)
    stack = []
    stackPaths = []
    fullZonePaths = []

    for line in lines:
        if 'END' in line:
            break

        indent = len(list(takewhile(is_tab, line)))
        newList = []

        stack[indent:] = newList
        stack[indent:] = [line.lstrip()]
        stackPaths[indent:] = [line.lstrip().split(';')]
        fullZonePaths.append(stackPaths[:])

    return fullZonePaths


@logger.catch()
# returns true if a special interactable teleporter was found and teleported to
async def teleportToInteractiveTeleportIfAvailable(p, currentZone, destinationZone, teleporterList):

    currentTeleporter = ''
    destinationZoneButtonNumber = ''
    for i in teleporterList:
        if i.strip() == 'END':
            break

        splitTeleporter = i.strip().split(';')

        if splitTeleporter[4].strip() == currentZone and splitTeleporter[4].strip() == destinationZone:
            currentTeleporter = ''
            destinationZoneButtonNumber = ''
            break

        elif splitTeleporter[4].strip() == currentZone.strip():
            currentTeleporter = i

        elif splitTeleporter[4].strip() == destinationZone.strip():
            destinationZoneButtonNumber = splitTeleporter[0].strip().split('_', 1)[1]

    # teleporter found in both current zone and destination zone.  This is our path
    if currentTeleporter != '' and destinationZoneButtonNumber != '':
        split = currentTeleporter.strip().split(';')

        await p.teleport(XYZ(float(split[1]), float(split[2]), float(split[3])), wait_on_inuse=True)
        await asyncio.sleep(.4)

        await interactiveTeleportToZone(p, int(destinationZoneButtonNumber) + 1)
        return True
    else:
        return False


async def read_control_checkbox_text(checkbox: Window) -> str:
    return await checkbox.read_wide_string_from_offset(616)

@logger.catch()
# trace back to the nearest spiral door and teleport to destinationWorld
async def goToNewWorld(p, destinationWorld):
    while not await p.is_in_npc_range():
        pass

    while await p.is_in_npc_range():
        await p.send_key(Keycode.X, 0.1)
        await asyncio.sleep(.4)

    await p.mouse_handler.activate_mouseless()

    # each worldList item (in-file name for a world) correlates to a zoneDoorOptions (in-file name for the buttons in the spiral door)
    worldList = ["WizardCity", "Krokotopia", "Marleybone", "MooShu", "DragonSpire", "Grizzleheim", "Celestia", "Wysteria", "Zafaria", "Avalon", "Azteca", "Khrysalis", "Polaris", "Arcanum", "Mirage", "Empyrea", "Karamelle", "Lemuria"]
    zoneDoorOptions = ["wbtnWizardCity", "wbtnKrokotopia", "wbtnMarleybone", "wbtnMooShu", "wbtnDragonSpire", "wbtnGrizzleheim", "wbtnCelestia", "wbtnWysteria", "wbtnZafaria", "wbtnAvalon", "wbtnAzteca", "wbtnKhrysalis", "wbtnPolaris", "wbtnArcanum", "wbtnMirage", "wbtnEmpyrea", "wbtnKaramelle", "wbtnLemuria"]
    zoneDoorNameList = ["Wizard City", "Krokotopia", "Marleybone", "MooShu", "DragonSpire", "Grizzleheim", "Celestia", "Wysteria", "Zafaria", "Avalon", "Azteca", "Khrysalis", "Polaris", "Arcanum", "Mirage", "Empyrea", "Karamelle", "Lemuria"]
    # user could be on any of the three pages when opening the world door depending on what their active quest is
    # switch all the way to the first page to standardize it
    for i in range(6):
        await p.mouse_handler.click_window_with_name('leftButton')
        await asyncio.sleep(0.2)

    option_window = await p.root_window.get_windows_with_name("optionWindow")

    assert len(option_window) == 1, str(option_window)

    for child in await option_window[0].children():
        if await child.name() == 'pageCount':
            pageCount = await child.maybe_text()
            pageCount = pageCount[8:-9]
            currentPage = pageCount.split('/', 1)[0]
            maxPage = pageCount.split('/', 1)[1]
            break

    # in case wizwalker misclicked / didn't click enough originally when resetting to page 1, ensure we are on page 1 (and if not click over again)
    while str(currentPage) != '1':
        await p.mouse_handler.click_window_with_name('leftButton')
        await asyncio.sleep(0.2)
        for child in await option_window[0].children():
            if await child.name() == 'pageCount':
                pageCount = await child.maybe_text()
                pageCount = pageCount[8:-9]
                currentPage = pageCount.split('/', 1)[0]

    worldIndex = worldList.index(destinationWorld)
    spiralGateName = zoneDoorNameList[worldIndex]

    isChildFound = False

    for i in range(int(maxPage)):
        for child in await option_window[0].children():
            if await child.name() in ['opt0', 'opt1', 'opt2', 'opt3']:
                name = await read_control_checkbox_text(child)
                if name == spiralGateName:
                    await p.mouse_handler.click_window_with_name(zoneDoorOptions[worldIndex])
                    await asyncio.sleep(.4)
                    await p.mouse_handler.click_window_with_name('teleportButton')
                    await p.wait_for_zone_change()

                    await p.mouse_handler.deactivate_mouseless()

                    # move away from the spiral door so we dont accidentally click on it again after teleporting later
                    await p.send_key(Keycode.W, 1.5)

                    isChildFound = True
                    break

        # correct world was not found - check the next page
        if not isChildFound:
            previousPage = currentPage
            loopCount = 0
            while currentPage == previousPage and loopCount < 30:
                loopCount += 1
                await p.mouse_handler.click_window_with_name('rightButton')

                # ensure that wizwalker didn't misclick and that we actually changed pages
                for child in await option_window[0].children():
                    if await child.name() == 'pageCount':
                        pageCount = await child.maybe_text()
                        pageCount = pageCount[8:-9]
                        currentPage = pageCount.split('/', 1)[0]
                #print(f"{await child.name()}: {await read_control_checkbox_text(child)}")

    # 4 buttons per page - menuButtonNum / 4 rounded up to nearest whole number equals the page that the button is on
    # worldIndex = worldList.index(destinationWorld)
    # actualButtonToClick = worldIndex + 1
    #
    # if actualButtonToClick > 4:
    #     pageNum = int(math.ceil(actualButtonToClick / 4)) - 1
    #     # click to correct page
    #     if pageNum > 0:
    #         for i in range(pageNum):
    #             await p.mouse_handler.click_window_with_name('rightButton')
    #             await asyncio.sleep(0.2)
    #
    #     # the actual button number on the page, considering that there are 4 buttons per page (0 -> 3)
    #     actualButtonToClick = ((actualButtonToClick) - ((pageNum) * 4))
    #
    #
    # await p.mouse_handler.click_window_with_name(zoneDoorOptions[worldIndex])
    # await asyncio.sleep(.4)
    # await p.mouse_handler.click_window_with_name('teleportButton')
    # await p.wait_for_zone_change()
    #
    # await p.mouse_handler.deactivate_mouseless()
    #
    # # move away from the spiral door so we dont accidentally click on it again after teleporting later
    # await p.send_key(Keycode.W, 1.5)

    bigStackDestinations = await createStack(destinationWorld)
    return bigStackDestinations

async def parseFile(fileName, worldName):
    file = open(fileName, "r")

    lines = file.readlines()
    start = 'WORLD - ' + worldName + '\n'

    try:
        indexStart = lines.index(start)
        removeIndex = indexStart + 1
        lines = lines[removeIndex:]
    except ValueError:
        await asyncio.sleep(.01)

    return lines


@logger.catch()
# from any zone in any world (excluding certain ones, such as aquila), travel to a destination zone
async def goToDestination(p, destinationZone, p1WorldName, bigStackDestinations, interactiveTeleportersOriginal):

    currentZone = await p.zone_name()
    currentWorld = currentZone.split('/', 1)[0]
    destinationWorld = destinationZone.split('/', 1)[0]

    pathToCurrentZoneStack = []
    pathToDestinationStack = []

    # user may not be in the correct world.  Find the nearest spiral door and teleport to the correct world
    if currentWorld != destinationWorld:
        p1ZoneNameNew = await p.zone_name()
        p1WorldNameNew = p1ZoneNameNew.split('/', 1)[0]

        # read list of unique locations, such as NPC locations and spiral door coordinates
        currentWorldObjectLocationsOriginal = await parseFile("traversalData/uniqueObjectLocations.txt", p1WorldNameNew)

        zoneDoor = ''
        zoneContainingZoneDoor = ''
        uniqueObjectIterator = 0
        while zoneDoor == '':
            split = currentWorldObjectLocationsOriginal[uniqueObjectIterator].split(';')
            if split[0] == 'ZONEDOOR':
                zoneDoor = split
                zoneContainingZoneDoor = split[4]

            uniqueObjectIterator += 1

        currentWorldStackDestinations = await createStack(currentWorld)

        # traverse to the first zone in the world (or wherever the zone door is located)
        logger.info(f'Returning to Spiral Door in: {zoneContainingZoneDoor}')
        await goToDestination(p, zoneContainingZoneDoor, currentWorld, currentWorldStackDestinations, interactiveTeleportersOriginal)

        # teleport to the zone door
        await p.teleport(XYZ(float(split[1]), float(split[2]), float(split[3])))
        await asyncio.sleep(2)

        # teleport to new world
        logger.info(f'User in wrong world - heading to {destinationWorld}')
        bigStackDestinations = await goToNewWorld(p, destinationWorld)
        p1WorldName = destinationWorld

    i = 0
    currentZone = await p.zone_name()
    if currentZone.strip() != destinationZone.strip():
        currentZoneInZoneMap = False
        while i < len(bigStackDestinations):
            if 'WORLD -' in bigStackDestinations[i][0][0]:
                break

            if str(bigStackDestinations[i][len(bigStackDestinations[i]) - 1][1]).strip() == currentZone.strip():
                currentZoneInZoneMap = True
                break
            i += 1

        if currentZoneInZoneMap == False:
            logger.info(f'Zone not found in zonemap - returning to Hub to reconnect')
            # these have to be hardcoded
            # if a player recalls to a dungeon, then attempts to return to the hub, they go through two zone changes: once to the hub, second back to the zone they were in before recalling
            # this makes it impossible to account for zone changes without having ridiculously long sleeps or hardcoded zone names
            worldHubsList = ['WizardCity/WC_Ravenwood_Teleporter', 'WizardCity/WC_Ravenwood', 'Krokotopia/KT_WorldTeleporter', 'Krokotopia/KT_Hub', 'Marleybone/Interiors/MB_WolfminsterAbbey', 'Marleybone/MB_Hub', 'DragonSpire/DS_Hub_Cathedral', 'MooShu/Interiors/MS_Teleport_Chamber', 'MooShu/MS_Hub', 'Celestia/CL_Hub', 'Wysteria/PA_Hub', 'Grizzleheim/GH_MainHub', 'Zafaria/ZF_Z00_Hub', 'Avalon/AV_Z00_Hub', 'Azteca/AZ_Z00_Zocalo', 'Khrysalis/KR_Z00_Hub', 'Polaris/PL_Z00_Walruskberg', 'Mirage/MR_Z00_Hub', 'Karamelle/KM_Z00_HUB', 'Empyrea/EM_Z00_Aeriel_HUB', 'Lemuria/LM_Z00_Hub']

            currentZone = await p.zone_name()
            while currentZone not in worldHubsList:
                await p.send_key(Keycode.END)
                await asyncio.sleep(.6)
                currentZone = await p.zone_name()

            await asyncio.sleep(3)
            currentZone = await p.zone_name()




    # for certain worlds, check if there is a teleporter between currentZone and destinationZone before traversing zones manually
    teleportedToZone = False
    if p1WorldName in ['Empyrea', 'Karamelle', 'Lemuria']:
        interactiveTeleportersOriginal = await parseFile('traversalData/interactiveTeleporters.txt', p1WorldName)
        teleportedToZone = await teleportToInteractiveTeleportIfAvailable(p, currentZone, destinationZone, interactiveTeleportersOriginal)

    # contains data about gates between zones
    linesOriginal = await parseFile("traversalData/gates_list.txt", p1WorldName)

    # iterate over the zonemap (bigStackDestinations) and find the index containing the player's current zone as well as index containing their destination
    if teleportedToZone == False:
        if currentZone != destinationZone.strip():
            pathToCurrentZoneStack = []
            pathToDestinationStack = []

            i = 0
            while i < len(bigStackDestinations) and (len(pathToCurrentZoneStack) == 0 or len(pathToDestinationStack) == 0):

                # if currentZone and destinationZone are the same, object is in the current zone, so only find the path to the destination
                if currentZone.strip() != destinationZone.strip():
                    if str(bigStackDestinations[i][len(bigStackDestinations[i]) - 1][1]).strip() == currentZone.strip():
                        pathToCurrentZoneStack = bigStackDestinations[i]

                if str(bigStackDestinations[i][len(bigStackDestinations[i]) - 1][1]).strip() == destinationZone.strip():
                    # copy full path to the destination
                    pathToDestinationStack = bigStackDestinations[i]

                i += 1




            # find all zones that the two paths (current and destination) have in common
            # the shortest path between two zones is found by returning to the most recent common path - so we throw out all common paths except the last one
            pathCurrentSplit = []
            pathDestinationSplit = []


            # in some worlds, zone maps are disconnected each other and connected through teleporters.  If we
            # haven't found a teleporter and cannot find a destination, trace back all the way to the start
            # and hopefully there will be a teleporter there
            if len(pathToDestinationStack) == 0 and len(pathToCurrentZoneStack) > 0:
                pathCurrentSplit.append(pathToCurrentZoneStack[0][0])

            a = 0
            while a < len(pathToCurrentZoneStack):
                pathCurrentSplit.append(pathToCurrentZoneStack[a][1])
                a += 1

            a = 0
            while a < len(pathToDestinationStack):
                pathDestinationSplit.append(pathToDestinationStack[a][1])
                a += 1

            commonZones = set(pathCurrentSplit).intersection(pathDestinationSplit)

            if len(pathCurrentSplit) > len(pathDestinationSplit):
                largerPath = 'pathCurrentSplit'
            elif len(pathDestinationSplit) > len(pathCurrentSplit):
                largerPath = 'pathDestinationSplit'
            else:
                largerPath = ''

            a = 0
            countRemoved = 0
            currZone = await p.zone_name()

            pathCurrentSplitModified = pathCurrentSplit
            pathDestinationSplitModified = pathDestinationSplit

            if largerPath == 'pathCurrentSplit' or largerPath == '':
                pathDestLength = len(pathDestinationSplit)
                while a < pathDestLength:
                    if pathCurrentSplit[a].strip() == pathDestinationSplit[a].strip():
                        if countRemoved == len(commonZones) - 1 and currZone != pathDestinationSplit[a].strip():
                            pathCurrentSplitModified = pathCurrentSplit[a + 1:]
                            pathDestinationSplitModified = pathDestinationSplit
                            countRemoved += 1

                        elif countRemoved == len(commonZones) - 1:
                            pathCurrentSplitModified = pathCurrentSplit[a + 1:]
                            pathDestinationSplitModified = pathDestinationSplit[a + 1:]
                            countRemoved += 1
                        else:
                            pathCurrentSplitModified = pathCurrentSplit[a + 1:]
                            pathDestinationSplitModified = pathDestinationSplit[a + 1:]
                            countRemoved += 1
                    a += 1
            else:
                pathCurrLength = len(pathCurrentSplit)

                while a < pathCurrLength:
                    if pathCurrentSplit[a].strip() == pathDestinationSplit[a].strip():
                        if countRemoved == len(commonZones) - 1 and currZone != pathDestinationSplit[a].strip():
                            pathCurrentSplitModified = pathCurrentSplit[a + 1:]
                            pathDestinationSplitModified = pathDestinationSplit
                            countRemoved += 1
                        elif countRemoved == len(commonZones) - 1:
                            pathCurrentSplitModified = pathCurrentSplit[a + 1:]
                            pathDestinationSplitModified = pathDestinationSplit[a + 1:]
                            countRemoved += 1
                        else:
                            pathCurrentSplitModified = pathCurrentSplit[a + 1:]
                            pathDestinationSplitModified = pathDestinationSplit[a + 1:]
                            countRemoved += 1

                        currReversed = list(reversed(pathCurrentSplitModified))
                    a += 1

            if len(pathCurrentSplitModified) > 0:
                pathCurrentSplitModified.pop()

            currReversed = list(reversed(pathCurrentSplitModified))

            pathToCurrentZoneStack = pathCurrentSplitModified
            pathToDestinationStack = pathDestinationSplitModified

            # traverse down the current path back to the common zone, then stop
            i = len(pathToCurrentZoneStack) - 1
            if len(pathToCurrentZoneStack) > 0:
                logger.info('Going back to most recent common zone')
                while i >= 0:

                    lines = iter(linesOriginal)
                    currZone = await p.zone_name()
                    gateFrom = pathToCurrentZoneStack[i].strip() + ';' + currZone

                    # zonemap zones are stored as zone gates, from beginning of world to end
                    returnGate = str(currZone) + ';' + str(pathToCurrentZoneStack[i].strip())
                    for q, s in enumerate(lines):
                        if s.strip() == 'END':
                            break

                        split = s.strip().split(';')
                        rg = split[4] + ';' + split[5]
                        if returnGate == rg:
                            splitString = s.strip().split(';')
                            x = splitString[1].strip()
                            y = splitString[2].strip()
                            z = splitString[3].strip()

                            await gateTypeDifferentiation(float(x), float(y), float(z), p, splitString[0])

                            currentZone = await p.zone_name()

                            if p1WorldName in ['Empyrea', 'Karamelle', 'Lemuria']:
                                await teleportToInteractiveTeleportIfAvailable(p, currentZone, destinationZone, interactiveTeleportersOriginal)
                            break

                    i -= 1

        # reached the most recent common zone between the current and destination zones
        # now traverse to the destination
        i = 0
        if len(pathToDestinationStack) != 0:
            logger.info(f'Heading to destination zone {destinationZone}')

        while i < len(pathToDestinationStack):

            teleportedToZone = False

            lines = iter(linesOriginal)
            currZone = await p.zone_name()
            zoneTo = pathToDestinationStack[i].strip()

            gateFrom = currZone + ';' + zoneTo

            if currZone.strip() != destinationZone.strip():
                for q, s in enumerate(lines):

                    if s.strip() == 'END':
                        break

                    split = s.strip().split(';')
                    gf = split[4] + ';' + split[5]
                    if gateFrom == gf:

                        splitString = s.strip().split(';')
                        x = splitString[1].strip()
                        y = splitString[2].strip()
                        z = splitString[3].strip()
                        await gateTypeDifferentiation(float(x), float(y), float(z), p, splitString[0])

                        currentZone = await p.zone_name()

                        if p1WorldName in ['Empyrea', 'Karamelle', 'Lemuria']:
                            await teleportToInteractiveTeleportIfAvailable(p, currentZone, destinationZone, interactiveTeleportersOriginal)

                        break

            i += 1

            # if we didn't find a gate matching the next destination zone, the two zones are probably connected by teleporters.  Try teleporting there
            zone = await p.zone_name()
            if zone != destinationZone.strip():
                if p1WorldName in ['Empyrea', 'Karamelle', 'Lemuria']:
                    logger.info(f'Physical gate not found.  Attempting interactive teleport')
                    await teleportToInteractiveTeleportIfAvailable(p, zone, zoneTo, interactiveTeleportersOriginal)


@logger.catch()
async def toZone(clients, destinationZone):
    currentZone = await clients[0].zone_name()
    worldName = currentZone.split('/', 1)[0]

    objectLines = await parseFile("traversalData/objectLocations.txt", worldName)
    interactiveTeleportersOriginal = await parseFile("traversalData/interactiveTeleporters.txt", worldName)
    uniqueObjectLocationsOriginal = await parseFile("traversalData/uniqueObjectLocations.txt", worldName)
    bigStackDestinations = await createStack(worldName)

    await asyncio.gather(*[goToDestination(p, destinationZone, worldName, bigStackDestinations, interactiveTeleportersOriginal) for p in clients])

    logger.info(f'reached destination zone: {destinationZone}.')

potion_ui_buy = [
    "fillallpotions",
    "buyAction",
    "btnShopPotions",
    "centerButton",
    "fillonepotion",
    "buyAction",
    "exit"
]

# Credit to Notfaj and Slackaduts for the majority of this function from UniversalBotSF
@logger.catch()
async def refillPotions(clients):
    await toZone(clients, 'WizardCity/WC_Ravenwood_Teleporter')

    # Walk to potion vendor
    for client in clients:

        try:
            await client.mouse_handler.activate_mouseless()
        except:
            logger.error('MOUSELESS ERROR')

        await client.goto(-0.5264079570770264, -3021.25244140625)
        await client.send_key(Keycode.W, 0.5)
        await client.wait_for_zone_change()
        await client.goto(11.836355209350586, -1816.455078125)
        await client.send_key(Keycode.W, 0.5)
        await client.wait_for_zone_change()
        await client.goto(-587.87927246093752, 404.43939208984375)
        await asyncio.sleep(1)
        await client.goto(-3965.254638671875, 1535.5472412109375)
        await asyncio.sleep(1)
        await client.goto(-4442.06005859375, 1001.5532836914062)
        await asyncio.sleep(1)
        while not await client.is_in_npc_range():
            await client.goto(-4442.06005859375, 1001.5532836914062)
        await client.send_key(Keycode.X, 0.1)
        await asyncio.sleep(1)
        # Buy potions
        while True:
            try:
                for i in "fillallpotions", "buyAction", "btnShopPotions", "centerButton", "fillonepotion", "buyAction", "exit":
                    await client.mouse_handler.click_window_with_name(i)
                    await asyncio.sleep(0.4)
            except ValueError:
                continue
            break

        try:
            await client.mouse_handler.deactivate_mouseless()
        except:
            logger.error('MOUSELESS ERROR')

        # Return
        await client.send_key(Keycode.PAGE_UP, 0.1)
        await client.wait_for_zone_change()


@logger.catch()
async def main(clientHandler):
    clientHandler.get_new_clients()
    clients = clientHandler.get_ordered_clients()
    p1, p2, p3, p4 = [*clients, None, None, None, None][:4]
    for i, p in enumerate(clients, 1):
        p.title = "p" + str(i)

    # Hook activation
    for p in clients:
        logger.info(f"[{p.title}] Activating Hooks")
        await p.activate_hooks()

    logger.info('Hooks activated!')

    
    try:
        await toZone(clients, 'WizardCity/WC_Streets/WC_OldeTown')
    finally:
        logger.info('Script complete')


async def run():
  clientHandler = ClientHandler()

  try:
    await main(clientHandler)
  except:
    import traceback

    traceback.print_exc()

  await clientHandler.close()


if __name__ == "__main__":
    asyncio.run(run())



from typing import *

from .combat_config_parser import PriorityLine

class BaseCombatBackend:
    def __init__(self, cast_time: float = 0.3):
        self.cast_time = cast_time
        self.combat = None

    def attach_combat(self, combat): # TODO: Import+type hint not working because of circular import
        self.combat = combat

    async def get_real_round(self, r: int) -> Optional[PriorityLine]:
        raise NotImplementedError()

    async def get_relative_round(self, r: int) -> Optional[PriorityLine]:
        raise NotImplementedError()

    async def handle_no_cards_given(self):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wizwalker',
 'wizwalker.extensions.wizsprinter',
 'wizwalker.extensions.wizsprinter.combat_backends',
 'wizwalker.extensions.wizsprinter.traversalData']

package_data = \
{"": ["*txt"]}

install_requires = \
['lark>=0.11.3,<0.12.0', 'wizwalker>=1.0.0']

setup_kwargs = {
    'name': 'wizsprinter',
    'version': '0.6.2',
    'description': 'A semi-official WizWalker extension',
    'long_description': 'Do not contact the author directly about this library - if you need to contact someone, contact the maintainer.\n\nWizSprinter is a semi-official extension to the `WizWalker library <https://github.com/StarrFox/wizwalker>`_.\n\nIt adds:\n\nWizSprinter (WizWalker extension):\n    - upgrade_clients\n    - extended get_new_clients\n    - extended remove_dead_clients\n    - extended get_ordered_clients\n\nSprintyClient (Client extension):\n    - better teleport\n    - get_base_entities_with_vague_name\n    - get_base_entities_with_behaviors\n    - get_health_wisps\n    - get_mana_wisps\n    - get_mobs\n    - find_safe_entities_from\n    - find_closest_of_entities\n    - find_closest_by_predicate\n    - find_closest_by_name\n    - find_closest_by_vague_name\n    - find_closest_health_wisp\n    - find_closest_mana_wisp\n    - find_closest_mob\n    - tp_to_closest_of\n    - tp_to_closest_by_name\n    - tp_to_closest_by_vague_name\n    - tp_to_closest_health_wisp\n    - tp_to_closest_mana_wisp\n    - tp_to_closest_mob\n    - calc_health_ratio\n    - calc_mana_ratio\n    - has_potion\n    - use_potion\n    - needs_potion\n    - use_potion_if_needed\n\nSprintyCombat (Implementation of a CombatHandler):\n    - CombatConfigProvider to parse a config file\n    - Full CombatHandler with some logic\n\nWizNavigator (Functionality related to traversing between zones):\n    - toZone\n    - refillPotions',
    'author': 'SirOlaf',
    'author_email': None,
    'maintainer': 'CowHunter0',
    'maintainer_email': 'cowhunter04@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import *

from wizwalker import XYZ, Keycode, MemoryReadError
from wizwalker.client import Client
from wizwalker.memory import DynamicClientObject, ClientObject


class SprintyClient(Client):
    async def remove_excluded_entities_from(self, entities: List[DynamicClientObject], excluded_ids: Set[int] = None) \
            -> List[DynamicClientObject]:
        if excluded_ids is not None and len(excluded_ids) > 0:
            res = []
            for e in entities:
                if excluded_ids is None or await e.global_id_full() not in excluded_ids:
                    res.append(e)
            return res
        return entities

    async def get_base_entity_list(self, excluded_ids: Set[int] = None) -> List[DynamicClientObject]:
        return await self.remove_excluded_entities_from(await super().get_base_entity_list(), excluded_ids)

    async def get_base_entities_with_predicate(self, predicate: Callable[[ClientObject], bool], excluded_ids: Set[int] = None):
        entities = []

        for entity in await self.get_base_entity_list(excluded_ids):
            if await predicate(entity):
                entities.append(entity)

        return entities

    async def get_base_entities_with_name(self, name: str, excluded_ids: Set[int] = None):
        return await self.remove_excluded_entities_from(await super().get_base_entities_with_name(name), excluded_ids)

    async def get_base_entities_with_vague_name(self, name: str, excluded_ids: Set[int] = None) -> List[DynamicClientObject]:
        async def _pred(e: ClientObject):
            if temp := await e.object_template():
                return name.lower() in (await temp.object_name()).lower()
            return False

        return await self.get_base_entities_with_predicate(_pred, excluded_ids)

    async def get_base_entities_with_behaviors(self, behaviors: List[str], excluded_ids: Set[int] = None):
        res = []
        for entity in await self.get_base_entity_list(excluded_ids):
            good = True
            try:
                inactive_behaviors = [await beh.read_type_name() for beh in await entity.inactive_behaviors()]
                for behavior in behaviors:
                    if behavior not in inactive_behaviors:
                        good = False
                        break
                if good:
                    res.append(entity)
            except (ValueError, MemoryReadError):
                continue
        return res

    async def get_health_wisps(self, excluded_ids: Set[int] = None) -> List[DynamicClientObject]:
        return await self.get_base_entities_with_vague_name("WispHealth", excluded_ids)

    async def get_mana_wisps(self, excluded_ids: Set[int] = None) -> List[DynamicClientObject]:
        return await self.get_base_entities_with_vague_name("WispMana", excluded_ids)

    async def get_mobs(self, excluded_ids: Set[int] = None) -> List[DynamicClientObject]:
        async def _pred(e: DynamicClientObject):
            try:
                behaviors = await e.inactive_behaviors()
                for b in behaviors:
                    if (await b.read_type_name()) == "NPCBehavior":
                        return await b.read_value_from_offset(288, "bool")
                return False
            except (ValueError, MemoryReadError):
                return False
        return await self.get_base_entities_with_predicate(_pred, excluded_ids)

    async def find_safe_entities_from(self, entities: List[DynamicClientObject], safe_distance: float = 700) \
            -> List[DynamicClientObject]:
        mob_positions = []
        for mob in await self.get_mobs():
            mob_positions.append(await mob.location())

        safe = []
        for entity in entities:
            pos = await entity.location()
            good = True
            for p in mob_positions:
                if p.distance(pos) < safe_distance:
                    good = False
                    break
            if good:
                safe.append(entity)

        return safe

    async def find_closest_of_entities(self, entities: List[DynamicClientObject], only_safe: bool = False) \
            -> Optional[DynamicClientObject]:
        closest = None
        smallest_dist = 0
        self_pos = await self.body.position()
        if only_safe:
            entities = await self.find_safe_entities_from(entities)
        for w in entities:
            dist = self_pos.distance(await w.location())
            if closest is None or dist < smallest_dist:
                smallest_dist = dist
                closest = w
        return closest

    async def find_closest_by_predicate(self, pred: Callable[[ClientObject], bool], only_safe: bool = False,
                                        excluded_ids: Set[int] = None) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_base_entities_with_predicate(pred, excluded_ids), only_safe)

    async def find_closest_by_name(self, name: str, only_safe: bool = False,
                                   excluded_ids: Set[int] = None) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_base_entities_with_name(name, excluded_ids), only_safe)

    async def find_closest_by_vague_name(self, name: str, only_safe: bool = False,
                                         excluded_ids: Set[int] = None) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_base_entities_with_vague_name(name, excluded_ids), only_safe)

    async def find_closest_health_wisp(self, only_safe: bool = False,
                                       excluded_ids: Set[int] = None) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_health_wisps(excluded_ids), only_safe)

    async def find_closest_mana_wisp(self, only_safe: bool = False,
                                     excluded_ids: Set[int] = None) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_mana_wisps(excluded_ids), only_safe)

    async def find_closest_mob(self, excluded_ids: Set[int] = None) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_mobs(excluded_ids), False)

    async def tp_to(self, entity: DynamicClientObject) -> bool:
        if entity is not None:
            try:
                await self.teleport(await entity.location())
                return True
            except (ValueError, MemoryReadError):
                return False
        return False

    async def tp_to_closest_of(self, entities: List[DynamicClientObject], only_safe: bool = False):
        if e := await self.find_closest_of_entities(entities, only_safe):
            await self.teleport(await e.location())
            return True
        return False

    async def tp_to_closest_by_name(self, name: str, only_safe: bool = False,
                                    excluded_ids: Set[int] = None) -> bool:
        if e := await self.find_closest_by_name(name, only_safe, excluded_ids):
            await self.teleport(await e.location())
            return True
        return False

    async def tp_to_closest_by_vague_name(self, name: str, only_safe: bool = False,
                                          excluded_ids: Set[int] = None) -> bool:
        if e := await self.find_closest_by_vague_name(name, only_safe, excluded_ids):
            await self.teleport(await e.location())
            return True
        return False

    async def tp_to_closest_health_wisp(self, only_safe: bool = False,
                                        excluded_ids: Set[int] = None) -> bool:
        return await self.tp_to_closest_of(await self.get_health_wisps(excluded_ids), only_safe)

    async def tp_to_closest_mana_wisp(self, only_safe: bool = False, excluded_ids: Set[int] = None) -> bool:
        return await self.tp_to_closest_of(await self.get_mana_wisps(excluded_ids), only_safe)

    async def tp_to_closest_mob(self, excluded_ids: Set[int] = None) -> bool:
        return await self.tp_to_closest_of(await self.get_mobs(excluded_ids), False)

    async def calc_health_ratio(self) -> float:
        """Simply returns current health divided by max health"""
        return await self.stats.current_hitpoints() / await self.stats.max_hitpoints()

    async def calc_mana_ratio(self) -> float:
        """Simply returns current mana divided by max mana"""
        return await self.stats.current_mana() / await self.stats.max_mana()

    async def needs_health(self, health_percent: int = 20) -> bool:
        return await self.calc_health_ratio() * 100 <= health_percent

    async def needs_mana(self, mana_percent: int = 10) -> bool:
        return await self.calc_mana_ratio() * 100 <= mana_percent

    async def has_potion(self) -> bool:
        return await self.stats.potion_charge() >= 1.0

    async def use_potion(self) -> bool:
        if await self.has_potion():
            try:
                await self.mouse_handler.click_window_with_name("btnPotions")
            except ValueError:
                return False
            return True
        return False

    async def needs_potion(self, health_percent: int = 20, mana_percent: int = 10) -> bool:
        return await self.needs_health(health_percent) or await self.needs_mana(mana_percent)

    async def use_potion_if_needed(self, health_percent: int = 20, mana_percent: int = 10) -> bool:
        """
        :param health_percent: How low health needs to be in percent
        :param mana_percent: How low mana needs to be in percent
        :return: False if it needed a potion but doesn't have any and True if it needed no potion or used one
        """
        if await self.needs_potion(health_percent, mana_percent):
            return await self.use_potion()
        return True



from typing import *

from lark import Transformer

from .combat_api import *

# TODO: bias for auto may be interesting


def get_sprinty_grammar():
    return r"""
            ?start: config
            config: line+
            
            line: round_specifier? move_config [(_pipe move_config)*]? _NEWLINE?
            
            move_config: move (_at target)?
            
            move: (move_pass | (spell enchant?))
            move_pass: "pass"
            
            spell: any_spell | words | string
            enchant: _open_bracket (words | string) _close_bracket
            
            target: (target_type | target_multi)
            target_type: target_self | target_boss | target_enemy | target_ally | target_aoe | target_named
            target_self: _spaced{"self"}
            target_boss: _spaced{"boss"}
            target_enemy: _spaced{"enemy"} (_open_paren INT _close_paren)?
            target_ally: _spaced{"ally"} (_open_paren INT _close_paren)?
            target_aoe: _spaced{"aoe"}
            target_named: words | string
            target_multi: _open_paren target_type [(_comma target_type)*]? _close_paren | target_type [(_comma target_type)*]?
            
            round_specifier: _newlines? "{" expression "}" _newlines?
            
            
            auto: _spaced{"auto"}
            
            any_spell: _spaced{"any"} _less_than spell_type (_and spell_type)* _greater_than
            spell_type: spell_damage | spell_aoe | spell_heal_self | spell_heal_other | spell_heal | spell_blade | spell_shield | spell_trap | spell_enchant
            spell_damage: _spaced{"damage"}
            spell_aoe: _spaced{"aoe"}
            spell_heal: _spaced{"heal"}
            spell_heal_self: spell_heal _spaced{"self"}
            spell_heal_other: spell_heal _spaced{"other"}
            spell_blade: _spaced{"blade"}
            spell_shield: _spaced{"shield"}
            spell_trap: _spaced{"trap"}
            spell_enchant: _spaced{"enchant"}
            
            expression: INT
            
            words: _newlines? word [word*] _newlines?
            word: NAME | ("0".."9")
            string: _newlines? ESCAPED_STRING _newlines?
            
            
            _open_paren: _spaced{"("}
            _close_paren: _spaced{")"}
            _open_bracket: _spaced{"["}
            _close_bracket: _spaced{"]"}
            _less_than: _spaced{"<"}
            _greater_than: _spaced{">"}
            _comma: _spaced{","}
            
            _at: _spaced{"@"}
            _pipe: _spaced{"|"}
            _and: _spaced{"&"}
            CR: "\r"
            LF: "\n"
            _NEWLINE: CR? LF
            _newlines: [_NEWLINE]*
            
            _spaced{tok}: _newlines? tok _newlines?
            
            
            %import common.INT
            %import common.CNAME -> NAME
            %import common.WS_INLINE
            %import common.ESCAPED_STRING
            
            
            %ignore WS_INLINE
        """


class TreeToConfig(Transformer):
    def spell(self, items):
        if type(items[0]) is not str:
            return TemplateSpell(items[0])
        else:
            name: str = items[0]
            if name.startswith("\""):
                return NamedSpell(name[1:-1], True)
            return NamedSpell(name, False)

    def enchant(self, items):
        return self.spell(items)

    def move_pass(self, items):
        return NamedSpell("pass")

    def move(self, items):
        return Move(*items)

    def move_config(self, items):
        if len(items) > 1 and type(items[1]) is not TargetType:
            t, n = items[1]
            if type(n) is str and n.startswith("\""):
                return MoveConfig(items[0], TargetData(t, n[1:-1], is_literal=True))
            return MoveConfig(items[0], TargetData(t, n))
        elif len(items) > 1:
            return MoveConfig(items[0], TargetData(items[1]))
        return MoveConfig(items[0])

    def line(self, items):
        if type(items[0]) is int:
            return PriorityLine(items[1:], items[0])
        return PriorityLine(items)

    def config(self, items):
        return CombatConfig(items)

    def target_self(self, _):
        return TargetType.type_self

    def target_boss(self, _):
        return TargetType.type_boss

    def target_enemy(self, items):
        if len(items) > 0:
            return TargetType.type_enemy, items[0]
        return TargetType.type_enemy

    def target_ally(self, items):
        if len(items) > 0:
            return TargetType.type_ally, items[0]
        return TargetType.type_ally

    def target_aoe(self, _):
        return TargetType.type_aoe

    def target_named(self, items):
        return TargetType.type_named, items[0]

    def target_type(self, items):
        return items[0]

    def target(self, items):
        return items[0]

    def any_spell(self, items):
        return items

    def spell_type(self, items):
        return items[0]

    def spell_damage(self, _):
        return SpellType.type_damage

    def spell_aoe(self, _):
        return SpellType.type_aoe

    def spell_heal_self(self, _):
        return SpellType.type_heal_self

    def spell_heal_other(self, _):
        return SpellType.type_heal_other

    def spell_heal(self, _):
        return SpellType.type_heal

    def spell_blade(self, _):
        return SpellType.type_blade

    def spell_shield(self, _):
        return SpellType.type_shield

    def spell_trap(self, _):
        return SpellType.type_trap

    def spell_enchant(self, _):
        return SpellType.type_enchant

    def expression(self, items):
        return items[0]

    def round_specifier(self, items):
        return items[0]

    INT = int

    def word(self, s):
        s, = s
        return s

    def words(self, items) -> str:
        res = ""
        for i in items:
            res += f" {i}"
        return res[1:]

    def string(self, items) -> str:
        res: str = items[0]
        res = res.encode("latin1").decode("unicode_escape")
        return res
